(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tenvir.ml,v 1.3 2001/02/21 10:14:24 bblanche Exp $ *)

open Types
open Typedtree
open Format
open Tsolver
open Tenv
open Ttype

module Envir =
struct

  type t = t_father * t_hierarchy

  and t_father = FSome of t | FNone

  let get_hierarchy env = Module (snd env)

  let startmodule (env : t) = 
     ((FSome env), emptyhierarchy)

  let add_module ident hierarchy (father, modulehierarchy) =
       (father, {env = modulehierarchy.env; 
         types = modulehierarchy.types;
         submod = Ident.add ident hierarchy
            modulehierarchy.submod;
         submod_types = modulehierarchy.submod_types})

  let add_module_type ident hierarchy (father, modulehierarchy) =
       (father, {env = modulehierarchy.env; 
         types = modulehierarchy.types;
         submod = modulehierarchy.submod;
         submod_types = Ident.add ident hierarchy
            modulehierarchy.submod_types})

  let update ident value ((father,modulehierarchy): t) =
      (father, {env = Ident.add ident value modulehierarchy.env;
                types = modulehierarchy.types;
                submod = modulehierarchy.submod;
                submod_types = modulehierarchy.submod_types})

  let add_type ident value ((father, modulehierarchy): t) =
      (father, {env = modulehierarchy.env;
                types = Ident.add ident value modulehierarchy.types;
                submod = modulehierarchy.submod;
                submod_types = modulehierarchy.submod_types})

  let rec replookup onelookup (father, hierarchy) =
     try
        onelookup hierarchy
     with
        Not_found -> match father with
           FSome f -> replookup onelookup f
         | FNone -> raise Not_found

  let rec lookup_module path env =
    try
     match path with
        Path.Pident(id) -> 
           (try
              replookup (fun hier -> Ident.find_same id hier.submod) env
           with Not_found ->
              Module(Allanal.toT(Compilenv.escape_info id)))
      | Path.Pdot(p,s,_) -> 
           let Module hier = lookup_module p env in 
              Ident.find_name s hier.submod
      | Path.Papply(p1,p2) ->
           let Functor(rand, body) = lookup_module p1 env in
              body
    with Not_found ->
        Message.debug 0 (fun () -> Printtyp.path path;
                                   print_string " not found");
        Module emptyhierarchy

  let lookup_ident path (env: t) =
     try 
        Some(match path with
           Path.Pident(id) -> replookup (fun hierarchy -> Ident.find_same 
              id hierarchy.env) env
         | Path.Pdot(p,s,n) -> let Module hierarchy = lookup_module p env in
              Ident.find_name s hierarchy.env
         | Path.Papply(_,_) -> Message.error "Envir.lookup" (fun () ->
             print_string "An identifier cannot be obtained by applying a functor"))
     with Not_found ->
        Message.debug 0 (fun () -> Printtyp.path path; 
                                   print_string " not found");
        None

  let lookup_type path (env: t) =
     try 
       match path with
           Path.Pident(id) -> replookup (fun hierarchy -> Ident.find_same 
              id hierarchy.types) env
         | Path.Pdot(p,s,n) -> let Module hierarchy = lookup_module p env in
              Ident.find_name s hierarchy.types
         | Path.Papply(_,_) -> Message.error "Envir.lookup_type" (fun () ->
             print_string "A type identifier cannot be obtained by applying a functor")
     with Not_found ->
        Message.error "Envir.lookup_type" (fun () -> print_string "Type identifier ";
                                   Printtyp.path path;
                                   print_string " not found")

  let lookup_module_type path env =
     try 
        match path with
           Path.Pident(id) -> replookup (fun hierarchy -> Ident.find_same 
             id hierarchy.submod_types) env
         | Path.Pdot(p,s,n) -> let Module hierarchy = lookup_module p env in
             Ident.find_name s hierarchy.submod_types
         | Path.Papply(_,_) -> Message.error "Envir.set_hierarchy_type" (fun () ->
             print_string "A module type identifier cannot be obtained by applying a functor")
     with Not_found ->
        Message.debug 0 (fun () -> print_string "Module type identifier ";
                                   Printtyp.path path;
                                   print_string " not found");
        Module emptyhierarchy

  let set_escape_info (_,modulehierarchy) =
      Compilenv.set_escape_info (Allanal.fromT modulehierarchy)

  let create() = (FNone, emptyhierarchy)

  let print (_, hier) = 
      let rec print_env hierarchy =
         match hierarchy with
             Module hier -> Ident.print_tbl IdentInfo.print_info hier.env;
                            print_modules hier.submod
           | Functor(_,body) -> print_env body
      and print_modules env =
         Ident.print_tbl (fun hierarchy -> print_env hierarchy) env
      in
         Message.debug 0 (fun () -> print_env (Module hier))

  let show_graphs (_, hier) = 
      let rec print_env hierarchy =
         match hierarchy with
             Module hier -> Ident.iter IdentInfo.show_graph hier.env;
                            print_modules hier.submod
           | Functor(_,body) -> print_env body
      and print_modules env =
         Ident.iter (fun _ hierarchy -> print_env hierarchy) env
      in
      let oc = open_out "envir.gra" in
      set_formatter_out_channel oc;
      print_env (Module hier);
      set_formatter_out_channel stdout;
      close_out oc


end
