(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Format
open Ttype
open Tsolver

let rec print_with_delimitors element separator list =
  match list with
    [] -> ()
  | [x] -> element x
  | (x::xs) -> 
      element x;
      print_string separator;
      print_with_delimitors element separator xs

let get = Graph.get

let print_sum ty =
  let rec annotate prio ty =
    if ty.b_mark = max_int then
      begin
	ty.b_mark <- 0;
	begin
          match (prio, ty.b_desc) with
            (_, BTgroup(decl,[])) -> 
	      Printtyp.ident decl.d_name
          | (_, BTgroup(decl,[idx,ty])) ->
	      annotate 0 ty;
	      print_string " ";
	      Printtyp.ident(decl.d_name)
          | (_, BTgroup(decl,types)) ->
	      if decl.d_modname = "" then
		begin
		  print_string "(";
		  print_with_delimitors (fun (idx, ty) -> 
		    annotate 2 ty) ", " types;
		  print_string ") "
		end
	      else
		begin
		  print_string decl.d_modname;
		  print_string "."
		end;
	      Printtyp.ident(decl.d_name);
          | (_, BTvar) -> print_string "'a"
          | (_, BTAll) -> print_string "All"
          | (_, BTabstract (decl, _, _)) ->
	      Printtyp.ident(decl.d_name);
          | (_, BTbadType) -> print_string "bad"
          | (0, _) ->
              (* This is not really a recursive call *)
              ty.b_mark <- max_int;
	      print_string "(";
	      annotate 2 ty;
	      print_string ")"
          | (2, BTarrow(idx1, idx2, ty1, ty2)) ->
	      annotate 1 ty1;
	      print_string " -> ";
	      annotate 2 ty2
          | (_, _) -> 
	      ty.b_mark <- max_int;
              annotate (prio-1) ty
              (* This is not really a recursive call *)
        end;
        ty.b_mark <- max_int
      end
    else print_string "rec"
  in
  annotate 1 ty

let get_name nt = "n" ^ (string_of_int nt.num)

let rec find_son i ty =
  if ty.b_mark = max_int then
    begin
      ty.b_mark <- 0;
      let res = 
      match ty.b_desc with
	BTvar | BTabstract _ | BTbadType | BTAll ->
	  Message.error "NewType.printgraph" (fun () -> print_string
	      "No son")
      |	BTarrow(_,idxdst,_,dst) -> 
	  if (!idxdst) == i then Some dst else
	  if (!idxdst) == -1 then find_son i dst else None
      |	BTgroup(_,l) -> find_list i l
      in
      ty.b_mark <- max_int;
      res
    end
  else
    None

and find_list i = function [] -> None
  | ((idx,ty)::l) -> if !idx == i then Some ty else 
    if !idx == -1 then
      match find_son i ty with
      	None -> find_list i l
      | l -> l
    else find_list i l

let cluster_count = ref 0

let rec printsons par_num ty nt =
  let name_nt = get_name nt in
  match nt.e_paramson with
    Some ntpar -> 
      print_string name_nt;
      print_string " [label=\"...->...\"];\n";
      print_string name_nt;
      print_string " -> ";
      print_string (get_name ntpar);
      print_string " [label=\"par";
      print_int par_num;
      print_string "\"];\nsubgraph cluster";
      incr cluster_count;
      print_int !cluster_count;
      print_string " {\n";
      let BTarrow(_,_,par,dst) = ty.b_desc in
      printsons 1 par ntpar;
      print_string "}\n";
      let res = get 0 nt in
      print_string name_nt;
      print_string " -> ";
      print_string (get_name res);
      print_string " [label=\"res\"];\nsubgraph cluster";
      incr cluster_count;
      print_int !cluster_count;
      print_string " {\n";
      printsons (par_num+1) dst res;
      print_string "}\n";
      print_string "{ rank = same; ";
      print_string (get_name ntpar);
      print_string "; ";
      print_string (get_name res);
      print_string "; }\n"
  | None -> 
  print_string name_nt;
  open_hbox();
  print_string " [label=\"";
  print_sum ty;
  print_string "\"";
  if nt.node_replacer = Escape then
    print_string ", color=red";
  print_string "];";
  close_box();
  print_newline();
  match nt.e_sons with
    Compact cptdesc -> ()
  | Expanded nsons ->
      for i = 0 to Array.length nsons - 1 do
	let Some tyi = find_son i ty in
	print_string name_nt;
	print_string " -> ";
	print_string (get_name nsons.(i));
	print_string ";\n";
	printsons 1 tyi nsons.(i)
      done

let rec printgraph nt =
  let name_nt = get_name nt in
  (* Prints edges *)
  match nt.e_sons with
    Compact cptdesc ->
      print_string name_nt;
      print_string " [shape = box];\n";
      (* Red edges = source edges *)
      Nset.iter (fun m ->
	print_string name_nt;
	print_string " -> ";
	print_string (get_name m);
	print_string " [color=red];\n") cptdesc.cpt_linksource;
      (* Green edges = son edges *)
      Nset.iter (fun m ->
	print_string name_nt;
	print_string " -> ";
	print_string (get_name m);
	print_string " [color=green];\n") cptdesc.cpt_linksons
  | Expanded nsons ->
      Nset.iter (fun m ->
	print_string name_nt;
	print_string " -> ";
	print_string (get_name m);
	print_string " [color=blue];\n") nt.norm;
  (* Prints edges of sons *)
  match nt.e_paramson with
    Some ntpar -> 
      printgraph ntpar;
      printgraph nsons.(0);
  | None -> 
      for i = 0 to Array.length nsons - 1 do
	printgraph nsons.(i)
      done

let graphprint name (ty,nt) =
  print_string "digraph ";
  print_string name;
  print_string " {\n";
  print_string "titlename [shape=box, label=\"";
  print_string name;
  print_string "\"];\n";
  printsons 1 ty nt;
  printgraph nt;
  print_string "}\n";
  print_flush()

let print mark (ty, nodetree) =
  let rec annotate prio just_print ty nodetree =
    if not just_print &&
      (let r = Graph.marked nodetree in
      Graph.unmark nodetree;
      r <= mark) then
      begin
        open_hovbox 0;
        print_string "H(";
        annotate 2 true ty nodetree;
        print_string ")";
        close_box()
      end
    else if ty.b_mark = max_int then
      begin
	ty.b_mark <- 0;
	begin
          match (prio, ty.b_desc) with
            (_, BTgroup(decl,[])) -> 
	      Printtyp.ident decl.d_name
          | (_, BTgroup(decl,[idx,ty])) ->
              open_hovbox 0;
	      print_string "[";
	      print_int (!idx);
	      print_string "]";
	      annotate 0 just_print ty (get (!idx) nodetree);
	      print_string " ";
	      Printtyp.ident(decl.d_name);
              close_box()
          | (_, BTgroup(decl,types)) ->
              open_hovbox 0;
	      print_string "(";
	      print_with_delimitors (fun (idx, ty) -> 
		print_string "[";
		print_int (!idx);
		print_string "]";
		annotate 2 just_print ty (get (!idx) nodetree)) ", " types;
	      print_string ") ";
	      Printtyp.ident(decl.d_name);
              close_box()
          | (_, BTvar) -> print_string "'a"
          | (_, BTAll) -> print_string "All"
          | (_, BTabstract (decl, _, _)) ->
	      Printtyp.ident(decl.d_name);
          | (_, BTbadType) ->
              print_string "bad"
          | (0, _) ->
                      (* This is not really a recursive call *)
              ty.b_mark <- max_int;
              open_hovbox 1;
	      print_string "(";
	      annotate 2 just_print ty nodetree;
	      print_string ")";
              close_box()
          | (2, BTarrow(idx1, idx2, ty1, ty2)) ->
              open_hovbox 0;
	      begin
		match nodetree.e_paramson with
		  None -> print_string "Rec"
		| Some param -> annotate 1 just_print ty1 param
	      end;
	      print_string " -> ";
	      annotate 2 just_print ty2 (get (!idx2) nodetree);
              close_box()
          | (_, _) -> ty.b_mark <- max_int;
              annotate (prio-1) just_print ty nodetree
                      (* This is not really a recursive call *)
        end;
        ty.b_mark <- max_int
      end
    else print_string "rec"
  in
  annotate 1 false ty nodetree

let iter_params f (_, nodetree) =
  let rec iterparamsrec nodetree =
    match nodetree.e_paramson with
      None -> ()
    | Some param -> f param; iterparamsrec (get 0 nodetree)
  in
  iterparamsrec nodetree

let tuple_name = {d_modname = ""; d_name = Ident.create "tuple"}
