(***********************************************************************)
(*                                                                     *)
(*              Fast Escape Analysis - Equations solver                *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tsolver.ml,v 1.44 2000/05/29 11:19:49 bblanche Exp $ *)

open Typedtree
open Format
open Ttype

module OrdNodes =
struct
  type t = Ttype.graph_t

  let compare x y = x.num - y.num
end

module Nset = Ordset.Make(OrdNodes)

module Graph =
struct

let curr_mark = ref 1

let incr_mark() = incr curr_mark

type t = Ttype.graph_t
(* {      (* Number to identify the node (for printing, sets...) *)
          num : int;

          (* How deep is the expression of the node in the abstract syntax
             tree *)
          temp_mark : int;

          (* First search : Is letstack possible ? *)
          mutable mark : int;
          mutable visited : int;

          (* Second search : Is letstack' possible ? *)
          mutable tail_mark : int;
          mutable tail_visited : int;

          (* Third search : Where does the parameters escape ? *)
          mutable type_replacer : (buildtype_t * newtype_t) list;
	  mutable node_replacer : nodereplacer_t;
          mutable replacer_visited : bool;

          (* Edges of the graph *)
          mutable pathEdge : t list;
          mutable sourcepathEdge : t list;
	  mutable norm : Nset.t; (* Normal edges *)
	  mutable rets : t list;
	  (* For each return, the node returned from *)
	  mutable starts : (t * (t * t) list * (t * t) list) list}
          (* For each start, the corresponding call node
             and a list of corresponding return nodes, for each one
                the return node and the exit node *)

          (* sons *)
          mutable e_sons : sons_t;
          mutable e_paramson : graph_t option}
*)

let count = ref 0

let newnode sons paramson = 
  incr count;
  { num = !count; mark = !curr_mark; tail_mark = !curr_mark; 
    temp_mark = !curr_mark;
    type_replacer = []; node_replacer = NoRep; replacer_visited = false;
    pathEdge = []; sourcepathEdge = []; visited = -1; tail_visited = -1;
    norm = Nset.empty; rets = []; starts = [];
    e_sons = sons; e_paramson = paramson}

let get_mark n = n.temp_mark

let array_create f n =
  if n = 0 then [| |] else
  let a = Array.create n (f 0) in
  for i = 1 to n-1 do
    a.(i) <- f i
  done;
  a

let expand_count = ref 0

let rec get_sons nt =
  match nt.e_sons with
    Expanded ntar -> ntar
  | Compact cptdesc -> 
      incr expand_count;
      if (!expand_count) mod 10000 = 0 then
	Message.debug 0 (fun () -> print_string "-- "; print_int !expand_count;
	  print_string " expansions");
      let ntar = array_create (fun i ->
        incr count;
	{ num = !count;
          temp_mark = nt.temp_mark;
	  mark = nt.mark;
	  visited = -1;
	  tail_mark = nt.tail_mark;
	  tail_visited = -1;
	  node_replacer = NoRep;
	  type_replacer = [];
	  replacer_visited = false;
	  pathEdge = [];
	  sourcepathEdge = [];
	  norm = Nset.empty;
	  rets = [];
	  starts = [];
	  e_sons = if Array.length cptdesc.cpt_sons.d_sons.(i).d_sons = 0 then
	    Expanded [| |]
	  else Compact({ cpt_pathEdge = [];
			 cpt_linksource = cptdesc.cpt_linksource;
			 cpt_linksons = Nset.empty;
			 cpt_rets = [];
			 cpt_starts = [];
			 cpt_sons = cptdesc.cpt_sons.d_sons.(i)});
	  e_paramson = None}
	  ) (Array.length cptdesc.cpt_sons.d_sons)
      in
      nt.e_sons <- Expanded ntar;
      for i = 0 to Array.length cptdesc.cpt_sons.d_sons - 1 do
	let soni = ntar.(i) in
	let sonslink = Nset.fold (fun x accu -> Nset.add (get i x) accu) cptdesc.cpt_linksons Nset.empty in
	soni.node_replacer <-
	   begin
	     match nt.node_replacer with
	       NoRep -> NoRep
	     | Escape -> Escape
	     | OneNode n -> OneNode (get i n)
	     | SeveralNodes l -> SeveralNodes (List.map (get i) l)
	   end;
	soni.pathEdge <- List.map (get i) cptdesc.cpt_pathEdge;
	soni.norm <- Nset.union sonslink cptdesc.cpt_linksource;
	soni.rets <- List.map (get i) cptdesc.cpt_rets;
	soni.starts <- List.map (fun (call,_,sonsrets) ->
	  (get i call, [], sonsrets)) cptdesc.cpt_starts;
	match soni.e_sons with
	  Expanded _ -> ()
	| Compact cpti ->
	    cpti.cpt_linksons <- sonslink;
	    cpti.cpt_rets <- soni.rets;
	    cpti.cpt_starts <- soni.starts;
	    cpti.cpt_pathEdge <- soni.pathEdge
      done;
      ntar

and get idx nodetree =
  if idx = -1 then nodetree else
  let sons = get_sons nodetree in
  if Array.length sons = 0 then nodetree else Array.get sons idx

(* Build the graph *)

let link n m = 
  if n.visited > -1 then 
    Message.debug 0 (fun () -> print_string "Node linked after visited");
  n.norm <- Nset.add m n.norm

let linksource n m =
  let rec linkrec n =
    link n m;
    match n.e_sons with
      Expanded nsons -> Array.iter linkrec nsons
    | Compact cptdesc ->
	cptdesc.cpt_linksource <- Nset.add m cptdesc.cpt_linksource
  in
  linkrec n

let rec linkdest n m =
  let rec linkrec m =
    link n m;
    Array.iter linkrec (get_sons m)
  in
  linkrec m

let rec linksons n m =
  if n != m then
    begin
      link n m;
      match n.e_sons with
	Expanded nsons ->
	  begin
	    let msons = get_sons m in
	    match (Array.length nsons, Array.length msons) with
              (0,0) -> ()
	    | (0,_) -> Array.iter (linkdest n) msons
	    | (_,0) -> Array.iter (fun n -> linksource n m) nsons
	    | (l,_) ->
              	for i = 0 to l-1 do
	    	  linksons nsons.(i) msons.(i)
              	done
	  end
      |	Compact cptdesc ->
	  cptdesc.cpt_linksons <- Nset.add m cptdesc.cpt_linksons
    end

let call paramcorresp paramsonscorresp resultcorresp resultsonscorresp =
  List.iter (fun (param_inst, param_gen) -> 
    param_inst.rets <- param_gen :: param_inst.rets
    ) paramcorresp;
  let rec write_paramsonscorresp (param_inst, param_gen) =
    param_inst.rets <- param_gen :: param_inst.rets;
    match param_inst.e_sons with
      Expanded isons ->
	let gsons = get_sons param_gen in
        for i = 0 to Array.length isons - 1 do
	  write_paramsonscorresp (isons.(i), gsons.(i))
	done
    | Compact cptdesc ->
	cptdesc.cpt_rets <- param_gen :: cptdesc.cpt_rets
  in
  List.iter write_paramsonscorresp paramsonscorresp;
  List.iter (fun (result_inst, result_gen) ->
    result_gen.starts <- (result_inst,paramcorresp,paramsonscorresp) :: result_gen.starts
      ) resultcorresp;
  let rec write_resultsonscorresp (result_inst, result_gen) =
    result_gen.starts <- (result_inst,paramcorresp,paramsonscorresp) :: result_gen.starts;
    match result_gen.e_sons with
      Expanded gsons ->
	let isons = get_sons result_inst in
        for i = 0 to Array.length isons - 1 do
	  write_resultsonscorresp (isons.(i), gsons.(i))
	done
    | Compact cptdesc ->
	cptdesc.cpt_starts <- (result_inst,[],paramsonscorresp) :: cptdesc.cpt_starts
        (* paramcorresp is useless *)
  in
  List.iter write_resultsonscorresp resultsonscorresp

exception ExpandedMet

let get_cpt_sons nodetree =
  match nodetree.e_sons with
    Expanded _ -> raise ExpandedMet
  | Compact cptdesc -> cptdesc

let rec mark n = 
  n.mark <- 0; n.tail_mark <- 0; n.node_replacer <- Escape;
  match n.e_sons with
    Expanded ntar -> Array.iter mark ntar
  | Compact cptdesc -> ()

let rec mark_tail n = 
  n.tail_mark <- 0;
  match n.e_sons with
    Expanded ntar -> Array.iter mark_tail ntar
  | Compact _ -> ()

let nostack = newnode (Expanded [| |]) None
let _ = mark nostack

exception Success

  let rec propagate n exit_p  = 
    if not (List.memq exit_p n.pathEdge) then
      begin
        n.pathEdge <- exit_p :: n.pathEdge;
        let success = ref false in
        List.iter (fun (call,ret,ret_sons) -> 
	  let call_update (r,e) = 
	    if (e == exit_p) then
	      begin
                success := true;
                if not (Nset.mem call r.norm) then
   		  begin
		    r.norm <- Nset.add call r.norm;
		    List.iter (fun exit_q -> propagate call exit_q)
		      r.pathEdge
		  end
	      end
	  in
          List.iter call_update ret;
	  let rec call_sons_update r e =
	    if (e == exit_p) then
	      begin
                success := true;
                if not (Nset.mem call r.norm) then
   		  begin
		    r.norm <- Nset.add call r.norm;
		    List.iter (fun exit_q -> propagate call exit_q)
		      r.pathEdge
		  end;
		raise Success
	      end
	    else
	      match e.e_sons with
		Expanded esons ->
		  if Array.length esons <> 0 then
		    let rsons = get_sons r in
		    for i = 0 to Array.length rsons - 1 do
		      call_sons_update  (Array.unsafe_get rsons i) (Array.unsafe_get esons i)
		    done
	      | Compact cptdesc -> ()
		     (* We won't find exit_p in these sons since they
			will be new ! *)
	  in
	  List.iter (fun (r,e) -> try 
	    call_sons_update r e
	  with Success -> ()) ret_sons) n.starts;
	if not !success then
	   (* CAREFUL ! I'm not sure about this. What should I do when a node
	      is at the same time the start node of a function and
	      a normal / return node ? This may happen since we both consider
	      the escapement of the parameters (for them, the result of the 
      	      function is labeled with start nodes), and the escapement of the 
      	      closure itself (result labeled with normal nodes) *)
       	  if n.node_replacer == Escape then
            exit_p.node_replacer <- Escape
          else begin
            Nset.iter (fun m -> propagate m exit_p) n.norm;
            List.iter (fun exit_q ->
              propagate exit_q exit_q;
	      if exit_q.node_replacer == Escape 
	      then exit_p.node_replacer <- Escape) n.rets
          end;
      end

let summUpdate n = propagate n n

let summUpdate = Profile.f1 "Graph.summUpdate" summUpdate

let rec sup n = function [] -> Misc.fatal_error "Graph.sup"
  | (a::l) -> if a == n then l else sup n l

let rec sourcepropagate n exit_p =
  if not (List.memq exit_p n.sourcepathEdge) then
    begin
      n.sourcepathEdge <- exit_p :: n.sourcepathEdge;
      try
      let esons = get_cpt_sons exit_p in
      let success = ref false in
      List.iter (fun (call,_,ret_sons) -> 
	let rec call_sons_update  r e =
	  if (e == exit_p) then
	    begin
              success := true;
              let rsons = get_cpt_sons r in
              if not (Nset.mem call rsons.cpt_linksource) then
   		begin
		  rsons.cpt_linksource <- Nset.add call rsons.cpt_linksource;
		  List.iter (fun exit_q -> sourcepropagate call exit_q)
		    rsons.cpt_pathEdge
		end;
              raise Success
	    end
	  else
	    match e.e_sons with
	      Expanded esons ->
		let rsons = get_sons r in
		for i = 0 to Array.length rsons - 1 do
		  call_sons_update  rsons.(i) esons.(i)
		done
	    | Compact cptdesc -> ()
		     (* We won't find exit_p in these sons since they
			will be new ! *)
	in
	List.iter (fun (r,e) -> 
          try
            call_sons_update r e
          with Success -> ()) ret_sons) n.starts;
      if not !success then
        if n.node_replacer == Escape then
	  exit_p.node_replacer <- Escape
        else begin
          List.iter (fun exit_q ->
	    summUpdate exit_q;
	    if exit_q.node_replacer == Escape 
	    then exit_p.node_replacer <- Escape) n.rets;
          Nset.iter (fun m -> sourcepropagate m exit_p) n.norm
        end
      with ExpandedMet ->
	n.sourcepathEdge <- sup exit_p n.sourcepathEdge;
	raise ExpandedMet
    end

let rec cptpropagate n exit_p  = 
  let nsons = get_cpt_sons n in
  if not (List.memq exit_p nsons.cpt_pathEdge) then
    begin
      nsons.cpt_pathEdge <- exit_p :: nsons.cpt_pathEdge;
      try
      let esons = get_cpt_sons exit_p in
      let success = ref false in
      List.iter (fun (call,_,ret_sons) -> 
	let rec call_sons_update  r e =
	  if (e == exit_p) then
	    begin
              success := true;
              let rsons = get_cpt_sons r in
              if not (Nset.mem call rsons.cpt_linksons) then
   		begin
		  rsons.cpt_linksons <- Nset.add call rsons.cpt_linksons;
		  List.iter (fun exit_q -> cptpropagate call exit_q)
		    rsons.cpt_pathEdge
		end;
              raise Success
	    end
	  else
	    match e.e_sons with
	      Expanded esons ->
		let rsons = get_sons r in
		for i = 0 to Array.length rsons - 1 do
		  call_sons_update  rsons.(i) esons.(i)
		done
	    | Compact cptdesc -> ()
		     (* We won't find exit_p in these sons since they
			will be new ! *)
	in
	List.iter (fun (r,e) -> 
          try
            call_sons_update r e
          with Success -> ()) ret_sons) nsons.cpt_starts;
      if not !success then
      	if n.node_replacer == Escape then
	  exit_p.node_replacer <- Escape
        else begin
          Nset.iter (fun m -> cptpropagate m exit_p) nsons.cpt_linksons;
	  Nset.iter (fun m -> sourcepropagate m exit_p) nsons.cpt_linksource; 
          List.iter (fun exit_q ->
            cptpropagate exit_q exit_q;
	    if exit_q.node_replacer == Escape 
	    then exit_p.node_replacer <- Escape) nsons.cpt_rets
        end
      with ExpandedMet ->
	nsons.cpt_pathEdge <- sup exit_p nsons.cpt_pathEdge;
	raise ExpandedMet
    end
    

let summcptUpdate n = cptpropagate n n

let summcptUpdate = Profile.f1 "Graph.summcptUpdate" summcptUpdate

let age = ref 0

(* First search : Is letstack possible ? *)

(* Scc algo by Tarjan *)
(* node -> old-node-list -> (min-mark, min-age, new-node-list) *)

let rec marked_scc (n : t) scc_list =
     if (n.mark = 0) then (0, max_int, scc_list) else
     if (n.visited > -1) then (n.mark, n.visited, scc_list) else
     begin
        incr age;
        n.visited <- !age;
	let scc_list = n :: scc_list in
	let (min_mark, min_age, scc_list) as accu =
	  if List.exists (fun exit_p ->
	    summUpdate exit_p;
	    exit_p.node_replacer == Escape) n.rets 
	  then (0, max_int, scc_list)
	  else Nset.fold marked_fold n.norm (n.mark, max_int, scc_list)
        in
	if (min_age >= n.visited) then
	  (* Entry node of a scc *)
	  let rec put_mark = function
               [] -> []
	     | (m::l) -> m.mark <- min_mark;
	                 m.visited <- max_int;
			 if m == n then l else put_mark l
          in (min_mark, max_int, put_mark scc_list)
        else
	  accu
     end
and marked_fold = fun n (min_mark, min_age, scc_list) ->
  let (min_mark', min_age', scc_list') = marked_scc n scc_list in
  (min min_mark min_mark', min min_age min_age', scc_list')

let marked n =
  age := 0;
  let (mark, _, _) = marked_scc n [] in
  mark

let rec unmark (n : t) =
     if n.visited = -1 then () else
     begin
        n.visited <- -1;
        (* summary edges already updated *)
        Nset.iter unmark n.norm;
     end

(* Second search : Is letstack' possible ? *)

let rec tail_marked_scc (n : t) scc_list =
     if (n.tail_mark = 0) then (0, max_int, scc_list) else
     if (n.tail_visited > -1) then (n.tail_mark, n.tail_visited, scc_list) else
     begin
        incr age;
        n.tail_visited <- !age;
	let scc_list = n :: scc_list in
	let (min_mark, min_age, scc_list) as accu =
	  if List.exists (fun exit_p ->
	    summUpdate exit_p;
	    exit_p.node_replacer == Escape) n.rets 
	  then (0, max_int, scc_list)
	  else Nset.fold tail_marked_fold n.norm
	      (n.tail_mark, max_int, scc_list)
        in
	if (min_age >= n.tail_visited) then
	  (* Entry node of a scc *)
	  let rec put_mark = function
               [] -> []
	     | (m::l) -> m.tail_mark <- min_mark;
	                 m.tail_visited <- max_int;
			 if m == n then l else put_mark l
          in (min_mark, max_int, put_mark scc_list)
        else
	  accu
     end
and tail_marked_fold = fun n (min_mark, min_age, scc_list) ->
  let (min_mark', min_age', scc_list') = tail_marked_scc n scc_list in
  (min min_mark min_mark', min min_age min_age', scc_list')

let tail_marked n =
  age := 0;
  let (mark, _, _) = tail_marked_scc n [] in
  mark


(* Third search : Where do the parameters escape ? *)

(* (1) To reduce the graph *)

let to_replace_by_node n x =
  match n.node_replacer with
    NoRep -> n.node_replacer <- OneNode x
  | Escape -> ()
  | OneNode y -> if x != y then n.node_replacer <- SeveralNodes [x;y]
  | SeveralNodes l -> if not (List.memq x l) then n.node_replacer <- SeveralNodes (x::l)

exception EscapeExc

let rec get_node_replacers_rec accu n = 
  if n.replacer_visited then accu else
  begin
    n.replacer_visited <- true;
    match n.node_replacer with
      OneNode m -> Nset.add m accu
    | SeveralNodes l -> List.fold_right Nset.add l accu
    | NoRep ->
	List.iter (fun exit_p ->
	  summUpdate exit_p;
	  if exit_p.node_replacer == Escape then raise EscapeExc) n.rets;
       	Nset.fold (fun n accu -> get_node_replacers_rec accu n) n.norm accu
    | Escape -> raise EscapeExc
  end

let rec cpt_get_node_replacers_rec accu n = 
  if n.replacer_visited then accu else
  begin
    n.replacer_visited <- true;
    match n.node_replacer with
      OneNode m -> Nset.add m accu
    | SeveralNodes l -> List.fold_right Nset.add l accu
    | NoRep ->
	let n = get_cpt_sons n in
	List.iter (fun exit_p ->
	  summcptUpdate exit_p;
	  if exit_p.node_replacer == Escape then
	    raise EscapeExc) n.cpt_rets;
       	Nset.fold (fun n accu -> 
	  cpt_get_node_replacers_rec accu n) n.cpt_linksons accu
    | Escape -> raise EscapeExc
  end

let rec cpt_get_source_node_replacers_rec accu n = 
  if n.replacer_visited then accu else
  begin
    n.replacer_visited <- true;
    match n.node_replacer with
      NoRep ->
	let n = get_cpt_sons n in
	List.iter (fun exit_p ->
	  summcptUpdate exit_p;
	  if exit_p.node_replacer == Escape then
	    raise EscapeExc) n.cpt_rets;
       	Nset.fold (fun n accu -> 
	  cpt_get_source_node_replacers_rec accu n) 
          n.cpt_linksons (Nset.fold (fun n accu -> get_node_replacers_rec accu n) 
	     n.cpt_linksource accu)
    | Escape -> raise EscapeExc
    | _ -> accu
  end

let rec cleanup_rep_visit n =
  if n.replacer_visited then
  begin
    n.replacer_visited <- false;
    (* Summary edges have already been updated by get_replacers_rec *)
    Nset.iter cleanup_rep_visit n.norm
  end

let rec cpt_cleanup_rep_visit n =
  if n.replacer_visited then
  begin
    n.replacer_visited <- false;
    (* Summary edges have already been updated by get_replacers_rec *)
    let n = get_cpt_sons n in
    Nset.iter cleanup_rep_visit n.cpt_linksource;
    Nset.iter (fun n -> cpt_cleanup_rep_visit n) n.cpt_linksons
  end

let get_node_replacers n = 
  try 
    let r = get_node_replacers_rec Nset.empty n in
    cleanup_rep_visit n;
    r
  with EscapeExc ->
    cleanup_rep_visit n;
    raise EscapeExc

let cpt_get_node_replacers n = 
  try 
    let r = cpt_get_node_replacers_rec Nset.empty n in
    cpt_cleanup_rep_visit n;
    r
  with x ->
    cpt_cleanup_rep_visit n;
    raise x

let cpt_get_source_node_replacers n = 
  try 
    let r = cpt_get_source_node_replacers_rec Nset.empty n in
    cpt_cleanup_rep_visit n;
    r
  with x ->
    cpt_cleanup_rep_visit n;
    raise x

(* (2) To use the reduced graph in instantiation *)

let to_replace_by_type n x = 
  n.type_replacer <- x :: n.type_replacer

let iter_type_replacers f gen_nt inst_nt = 
  Nset.iter (fun n ->
    List.iter f n.type_replacer;
    match n.node_replacer with
      NoRep -> ()
    | Escape -> Misc.fatal_error "Graph.iter_type 2"
    | OneNode m -> linksource inst_nt m
    | SeveralNodes l -> List.iter (linksource inst_nt) l
  ) gen_nt.norm

let copy_link gen_nt inst_nt =
  begin
  match gen_nt.node_replacer with
    Escape -> raise EscapeExc
  | NoRep -> ()
  | _ -> Misc.fatal_error "Graph.copy_link 1"
  end;
  Nset.iter (fun x ->
    List.iter (fun (_, (_, inst_res_nt)) -> link inst_nt inst_res_nt)
      x.type_replacer;
    match x.node_replacer with
      NoRep -> ()
    | Escape -> Misc.fatal_error "Graph.copy_link 2"
    | OneNode m -> link inst_nt m
    | SeveralNodes l -> List.iter (link inst_nt) l
  ) gen_nt.norm

let cpt_copy_link f cptdesc inst_nt =
  Nset.iter (fun n -> 
    List.iter f n.type_replacer;
    match n.node_replacer with
      NoRep -> ()
    | Escape -> Misc.fatal_error "Graph.cpt_copy_link 1"
    | OneNode m -> linksons inst_nt m
    | SeveralNodes l -> List.iter (linksons inst_nt) l
  ) cptdesc.cpt_linksons;
  Nset.iter (fun n ->
    List.iter (fun (_,(_,inst_res_nt)) -> linksource inst_nt inst_res_nt)
      n.type_replacer;
    match n.node_replacer with
      NoRep -> ()
    | Escape -> Misc.fatal_error "Graph.cpt_copy_link 2"
    | OneNode m -> linksource inst_nt m
    | SeveralNodes l -> List.iter (linksource inst_nt) l
  ) cptdesc.cpt_linksource

let rec cleanup_replace n =
  if n.node_replacer != Escape then n.node_replacer <- NoRep;
  n.type_replacer <- [];
  match n.e_sons with
    Expanded nsons -> Array.iter cleanup_replace nsons
  | Compact cptdesc -> ()

(* Copy the graph (useful for inlining) *)

let rec get_sons2 nt =
  match nt.e_sons with
    Expanded ntar -> ntar
  | Compact cptdesc -> 
      incr expand_count;
      if (!expand_count) mod 10000 = 0 then
	Message.debug 0 (fun () -> print_string "-- "; print_int !expand_count;
	  print_string " expansions");
      let ntar = array_create (fun i ->
        incr count;
	{ num = !count;
          temp_mark = nt.temp_mark;
	  mark = nt.mark;
	  visited = -1;
	  tail_mark = nt.tail_mark;
	  tail_visited = -1;
	  node_replacer = NoRep;
	  type_replacer = [];
	  replacer_visited = false;
	  pathEdge = [];
	  sourcepathEdge = [];
	  norm = Nset.empty;
	  rets = [];
	  starts = [];
	  e_sons = if Array.length cptdesc.cpt_sons.d_sons.(i).d_sons = 0 then
	    Expanded [| |]
	  else Compact({ cpt_pathEdge = [];
			 cpt_linksource = cptdesc.cpt_linksource;
			 cpt_linksons = Nset.empty;
			 cpt_rets = [];
			 cpt_starts = [];
			 cpt_sons = cptdesc.cpt_sons.d_sons.(i)});
	  e_paramson = None}
	  ) (Array.length cptdesc.cpt_sons.d_sons)
      in
      nt.e_sons <- Expanded ntar;
      for i = 0 to Array.length cptdesc.cpt_sons.d_sons - 1 do
	let soni = ntar.(i) in
	let sonslink = Nset.fold (fun x accu -> Nset.add (get2 i x) accu) cptdesc.cpt_linksons Nset.empty in
	(*soni.node_replacer <-
	   begin
	     match nt.node_replacer with
	       Escape -> Escape
	     | _ -> NoRep
	   end;*)
	soni.pathEdge <- List.map (get2 i) cptdesc.cpt_pathEdge;
	soni.norm <- Nset.union sonslink cptdesc.cpt_linksource;
	soni.rets <- List.map (get2 i) cptdesc.cpt_rets;
	soni.starts <- List.map (fun (call,_,sonsrets) ->
	  (get2 i call, [], sonsrets)) cptdesc.cpt_starts;
	match soni.e_sons with
	  Expanded _ -> ()
	| Compact cpti ->
	    cpti.cpt_linksons <- sonslink;
	    cpti.cpt_rets <- soni.rets;
	    cpti.cpt_starts <- soni.starts;
	    cpti.cpt_pathEdge <- soni.pathEdge
      done;
      ntar

and get2 idx nodetree =
  if idx = -1 then nodetree else
  let sons = get_sons2 nodetree in
  if Array.length sons = 0 then nodetree else Array.get sons idx

let copylist = ref []

let copy inc n =
  let rec copyrec n =
  (* inc = increment to apply to marks *)
    match n.node_replacer with
      OneNode m -> m
    | NoRep | Escape ->
    	copylist := n :: !copylist;
	(*
	if List.exists (fun exit_p ->
	  summUpdate exit_p;
	  exit_p.node_replacer == Escape) n.rets 
	then begin
	  n.node_replacer <- Escape;
	  n.mark <- 0;
	  n.tail_mark <- 0
	end;
	*)
	List.iter (fun exit_p -> summUpdate exit_p) n.rets;
	incr count;
      	let ncopy = 
	  { num = !count;
	    mark = if n.mark = 0 then 0 else n.mark + inc;
	    tail_mark = if n.tail_mark = 0 then 0 else n.tail_mark + inc;
	    temp_mark = n.temp_mark + inc;
	    type_replacer = [];
	    node_replacer = n.node_replacer;
	    replacer_visited = false;
	    pathEdge = [];
	    sourcepathEdge = [];
	    visited = -1;
	    tail_visited = -1;
	    norm = Nset.empty;
	    rets = [];
	    starts = [];
	    e_sons = Expanded [| |];
	    e_paramson = None
	  } in
    	n.node_replacer <- OneNode ncopy;
	ncopy.norm <- Nset.fold (fun n accu -> Nset.add (copyrec n) accu) n.norm Nset.empty;
	ncopy.e_sons <- 
	    begin
	      match n.e_sons with
	      	Expanded nsons -> Expanded (Array.map copyrec nsons)
	      | Compact cptdesc -> 
		  try
		    List.iter (fun exit_p -> 
		      summcptUpdate exit_p(*;
		      if exit_p.node_replacer == Escape
		      then begin 
			ncopy.node_replacer <- Escape;
			ncopy.mark <- 0;
			ncopy.tail_mark <- 0
		      end*)) cptdesc.cpt_rets;
		    Compact
		    { cpt_pathEdge = [];
		      cpt_linksource = Nset.fold (fun n accu -> Nset.add (copyrec n) accu) cptdesc.cpt_linksource Nset.empty;
		      cpt_linksons = Nset.fold (fun n accu -> Nset.add (copyrec n) accu) cptdesc.cpt_linksons Nset.empty;
		      cpt_rets = [];
		      cpt_starts = [];
		      cpt_sons = cptdesc.cpt_sons}
		  with ExpandedMet ->
		    (* If we meet an expanded node, we force expansion *)
		    Expanded (Array.map copyrec (get_sons2 n))
	    end;
	ncopy.e_paramson <- 
	   begin
	     match n.e_paramson with
	 	None -> None
	      | Some param -> Some (copyrec param)
	   end;
    	ncopy
  in
  (* do not copy nostack, since we compare its address with == to forbid
     stack allocation *)
  if (n == nostack)
  then nostack
  else copyrec n

let rec cleanupallcopies () = 
  List.iter (fun n ->
    match n.node_replacer with
      OneNode m -> n.node_replacer <- m.node_replacer
    | _ -> ()) !copylist;
  copylist := []
  

end
