(***********************************************************************)
(*                                                                     *)
(*              Fast Escape Analysis - Equations solver                *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: message.ml,v 1.2 2001/03/08 10:48:59 bblanche Exp $ *)

open Format
open Times

exception ErrorException

let debug_level = ref(1)

type time = float * float

let currentState: (string * int * time) list ref = ref([])
let phaseNumber: int ref = ref(1)

let push(entry,stack) =
    stack := entry :: !stack

let pop(stack) =
    let (top::rest) = !stack
    in
	begin
	    stack := rest;
	    top
	end

let getUserTime(): time = 
    let t = times()
    in
	(t.tms_utime, t.tms_stime)


    let reset() =
	(currentState := [];
	 phaseNumber := 1)

    let rec printPhaseNumber l =
	match l with
		[] -> ()
	      | ((_, number, _)::[]) -> print_int number
	      | ((_, number, _)::rest) -> printPhaseNumber rest;
                                          print_string ".";
                                          print_int number

    let printTime((usr',sys'), (usr,sys)) =
	let usr = (usr' -. usr) in
	let sys = (sys' -. sys) in
	    print_float usr;
            print_string "u + ";
            print_float sys;
            print_string "s"


    let stars() = String.make (3+List.length(!currentState)) '*'

    let send kind whom description =
        open_hovbox 0;
	print_string (stars());
        print_string " Phase ";
	printPhaseNumber !currentState;
	print_string " (";
	if (whom <> "") & (!currentState <> []) then
        begin
	   let (x, _, _):: _ = !currentState
	   in
	      print_string x;
              print_string "."
        end;
	print_string whom;
	print_string ") ";
        print_string kind;
        print_string ":";
        print_break 1 0;
	description();
        print_newline();
        close_box()

    let phasestart name =
        let stars = stars() in
	let stackEntry = (name, !phaseNumber, getUserTime()) in
	    push(stackEntry, currentState);
            open_hovbox 0;
	    phaseNumber := 1;
	    print_string stars;
            print_string " Phase ";
	    printPhaseNumber !currentState;
	    print_string " (";
            print_string name;
            print_string ")";
            print_newline();
            close_box()

    let phaseend name =
	let endTime = getUserTime() in
	let phase = !currentState in
	let (name', oldNumber, startTime) = pop(currentState) in
	    if (name <> name') then
		(output_string Pervasives.stderr ("*** currently in phase " ^ name' ^ " I got a PhaseEnd(" ^ name ^ ")\n");
		 raise ErrorException)
	    else
		begin
		    phaseNumber := oldNumber + 1;
                    open_hovbox 0;
		    print_string (stars());
                    print_string " Phase ";
		    printPhaseNumber phase;
		    print_string " (";
                    print_string name;
                    print_string ") took ";
		    printTime(endTime,startTime);
                    print_newline();
                    close_box()
		end

   let debug level description =
	if (level <= !debug_level) then
        begin
           open_hovbox 0;
	   description();
           print_newline();
           close_box()
        end

   let error whom description =
	send "ERROR" whom description;
	raise ErrorException

