(* $Id: pretty.ml,v 1.1 1998/02/15 18:16:07 blanchet Exp $ *)

(**** ML Programs from the book

  ML for the Working Programmer
  by Lawrence C. Paulson, Computer Laboratory, University of Cambridge.
  (Cambridge University Press, 1991)

Copyright (C) 1991 by Cambridge University Press.
Permission to copy without fee is granted provided that this copyright
notice and the DISCLAIMER OF WARRANTY are included in any copy.

DISCLAIMER OF WARRANTY.  These programs are provided `as is' without
warranty of any kind.  We make no warranties, express or implied, that the
programs are free of error, or are consistent with any particular standard
of merchantability, or that they will meet your requirements for any
particular application.  They should not be relied upon for solving a
problem whose incorrect solution could result in injury to a person or loss
of property.  If you do use the programs or functions in such a manner, it
is at your own risk.  The author and publisher disclaim all liability for
direct, incidental or consequential damages resulting from your use of
these programs or functions.
****)


(*** Pretty printing.  See Oppen (1980).  From Chapter 8.  ***)

  type t = 
      Block of t list * int * int 	(*indentation, length*)
    | String of string
    | Break of int			(*length*)
    | Nl
    | Nl'				(* consecutive Nl's are collapsed to one,
					   unlike Nl.
					 *)

  let rec makestring(x) =
      match x with
      (String s) -> s
    | (Break n) -> if n > 0 then " " ^ (makestring(Break(n-1))) else ""
    |(Block([], _, _)) -> ""
    |(Block(exp::exps, i, j)) -> makestring(exp) ^ makestring(Block(exps, i, j))
    |(Nl) -> "\n"
    |(Nl') -> "\n"

  (*Add the lengths of the expressions until the next Break; if no Break then
    include "after", to account for text following this block. *)

  let rec breakdist(x) =
      match x with
	  (Block(_,_,len)::sexps, after) -> len + breakdist(sexps, after)
    | (String s :: sexps, after) -> String.length s + breakdist (sexps, after)
    | (Break _ :: sexps, after) -> 0
    | (Nl :: sexps, after) ->0
    | (Nl' :: sexps, after) ->0
    | ([], after) ->after

  let rec pr ((os: out_channel), (sexp: t), (margin: int)) =
   let space = ref margin in
   let rec blanks n =
       if n = 0 then
	   ()
       else
	   (output_string os " ";  space := !space - 1; 
	    blanks(n-1)) in

   let newline () = (output_string os "\n";  space := margin) in

   let rec printing(sexps, (blockspace: int), after) =
       match sexps with
	   [] -> ()
	 | (sexp::sexps) ->
	       begin
		   match sexp with
		    Block(bsexps,indent,len) ->
			printing(bsexps, !space-indent, breakdist(sexps,after))
		  | String s -> (output_string os s;   space := !space - String.length s)
		  | Break len -> 
		    if len + breakdist(sexps,after) <= !space then
			blanks len
		    else
			(newline();  blanks(margin-blockspace))
		  | Nl -> (newline(); blanks(margin-blockspace))
		  | Nl' ->
			if (!space >= blockspace) then
			    ()
			else
			    (newline(); blanks(margin-blockspace))
	       end;
	       printing (sexps, blockspace, after)
   in
       printing([sexp], margin, 0);
       newline();
       flush(os)

  let rec length(x) =
      match x with
	  (Block(_,_,len)) -> len
    | (String s) -> String.length s
    | (Break len) -> len
    | (Nl) -> 0
    | (Nl') -> 0

  let str(x) = String(x) and brk(x) = Break(x) and nl = Nl and nl' = Nl'

  let blo (indent,sexps) =
      let rec sum(sexps, k) =
	  match sexps with
	      [] -> k
	    | (sexp::sexps) -> sum(sexps, length sexp + k)
      in
	  Block(sexps,indent, sum(sexps,0))

  let rec insertDelimitors(d, l) =
      match l with
	  [] -> []
	| [x] -> [x]
	| x::l -> x::(d@insertDelimitors(d,l))

  let rec duplicateString(string, n) =
      if (n = 0) then ""
      else 
	  if (n = 1) then string
	  else string ^ duplicateString(string, n-1)

  let pad(string, left_padding, right_padding) =
      let left_padding = max left_padding 0 in
      let right_padding = max right_padding 0
      in
	  duplicateString(" ", left_padding) ^ string ^
	  duplicateString(" ", right_padding)

  let padLeft(string, width) =
      let left_padding = width - String.length(string) in
      let right_padding = 0
      in
	  pad(string, left_padding, right_padding)

  let padRight(string, width) =
      let left_padding = 0 in
      let right_padding = width - String.length(string)
      in
	  pad(string, left_padding, right_padding)

  let padCenter(string, width) =
      let delta = width - String.length(string) in
      let left_padding = (delta / 2) in
      let right_padding = (delta / 2) + (delta mod 2)
      in
	  (duplicateString(" ", left_padding)) ^ string ^ (duplicateString(" ", right_padding))



