(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                           Pretransformation                         *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Types
open Typedtree
open Format
open Asttypes

exception Too_big
exception Not_higher_order

let tmp_counter = ref(0)

let to_trivial expr =
       incr(tmp_counter);
       let ident = Ident.create("%t" ^ string_of_int(!tmp_counter)) in
       let ident_as_pattern =
          {pat_desc = Tpat_var(ident);
           pat_loc = Location.none;
           pat_type = expr.exp_type} in
       let ident_as_expr = {exp_desc = Texp_ident(Path.Pident(ident),
                               {val_type = expr.exp_type;
                                val_kind = Val_reg});
                            exp_loc = expr.exp_loc;
                            exp_type = expr.exp_type;
                            exp_env = expr.exp_env} in
        ((ident_as_pattern, expr), ident_as_expr)

let rec create_pats =
   function [] -> ([], [])
         | (a::l) -> let (bindings, args) = create_pats l in
                      match a.exp_desc with
                         Texp_ident(_,_) | Texp_constant(_) ->
                            (bindings, a::args)
                       | _ ->
                         let (binding, arg) = to_trivial a
                         in (binding :: bindings, arg::args)

let rec list_combine_add l1 l2 =
   match (l1, l2) with
     ([], l) -> create_pats l
   | (a::l, b::m) -> let (l1, l2) = list_combine_add l m in
                        ((a,b)::l1, l2)
   | (_,_) -> invalid_arg "list_combine_add"

let rec list_combine_cut l1 l2 =
   match (l1, l2) with
     ([], l) -> ([], l)
   | (a::l, b::m) -> let (l1, l2) = list_combine_cut l m in
                        ((a,b)::l1, l2)
   | (_,_) -> invalid_arg "list_combine_cut"

type inline_info = (int * int * Typedtree.expression) Ident.tbl

let inline_sup_info = ref (Ident.empty : inline_info)

let indesc e desc = 
  {exp_desc = desc;
   exp_loc = e.exp_loc;
   exp_type = e.exp_type;
   exp_env = e.exp_env
   }

(* Computes the size of an expression *)

let rec taille taillemax e = if taillemax = 0 then
   raise Too_big
else match e.exp_desc with
   Texp_ident(_,_) -> taillemax-1
 | Texp_constant(_) -> taillemax-1
 | Texp_let(_,_, bindings, body) -> taille (taillebind (taillemax-1) bindings) body
 | Texp_function(clauses) -> taillebind (taillemax-1) clauses
 | Texp_apply(rator, rands) -> taille (List.fold_left taille (taillemax-1) rands) rator
 | Texp_match(body, clauses) -> taille (taillebind (taillemax-1) clauses) body
 | Texp_try(body, clauses) -> taille (taillebind (taillemax-1) clauses) body
 | Texp_tuple(rands) -> List.fold_left taille (taillemax-1) rands
 | Texp_construct(_, rands) -> List.fold_left taille (taillemax-1) rands
 | Texp_record(fielddesc) -> let (_, rands) = List.split fielddesc in
      List.fold_left taille (taillemax-1) rands
 | Texp_field(body, _) -> taille (taillemax-1) body
 | Texp_setfield(lhs, _, rhs) -> taille (taille (taillemax-1) lhs) rhs
 | Texp_array(rands) -> List.fold_left taille (taillemax-1) rands
 | Texp_ifthenelse(e1, e2, e3) -> let t = taille (taille (taillemax-1) e1) e2 in
      (match e3 with
          None -> t
        | Some e -> taille t e)
 | Texp_sequence(e1, e2) -> taille (taille (taillemax-1) e1) e2
 | Texp_while(cond, body) -> taille (taille (taillemax-1) cond) body
 | Texp_for(_, first, last, _, body) -> taille (taille (taille (taillemax-1) first) last) body
 | Texp_when(cond, body) -> taille (taille (taillemax-1) cond) body 
 | Texp_send(exp,_) -> taille (taillemax-1) exp
 | Texp_new _ -> taillemax-1
 | Texp_instvar _ -> taillemax-1
 | Texp_setinstvar (_,_,exp) -> taille (taillemax-1) exp
 | Texp_override (_, rands) -> List.fold_left taille (taillemax-1) (List.map snd rands)

and taillepat taillemax pat = if taillemax = 0 then
   raise Too_big
else match pat.pat_desc with
   Tpat_any -> taillemax-1
 | Tpat_var(_) -> taillemax-1
 | Tpat_alias(pat, _) -> taillepat (taillemax-1) pat
 | Tpat_constant(_) -> taillemax-1
 | Tpat_tuple(patlist) ->  List.fold_left taillepat (taillemax-1) patlist
 | Tpat_construct(_, patlist) -> List.fold_left taillepat (taillemax-1) patlist
 | Tpat_record(fieldpat) -> let (_, patlist) = List.split fieldpat in
       List.fold_left taillepat (taillemax-1) patlist
 | Tpat_or(pat1, pat2) -> taillepat (taillepat (taillemax-1) pat1) pat2
   
and taillebind taillemax bindings = if taillemax = 0 then
   raise Too_big
else match bindings with
   [] -> taillemax
 | ((pat, body)::bindings') -> taillebind (taillepat (taille 
         (taillemax-1) body) pat) bindings'

(* Instanciates types *)
let generic_level = (-1)
let not_copy_level = (-1) (* Not used by the type checker *)

let saved_desc = ref []
  (* Saved association of generic nodes with their description. *)


(**** Object field manipulation. Copied form ctype.ml ****)

let flatten_fields ty =
  let rec flatten l ty =
    let ty = Ctype.repr ty in
    match ty.desc with
      Tfield(s, ty1, ty2) ->
        flatten ((s, ty1)::l) ty2
    | _ ->
        (l, ty)
  in
    let (l, r) = flatten [] ty in
      (List.rev l, r)

let build_fields =
  List.fold_right (fun (s, ty1) ty2 -> Ctype.newty (Tfield(s, ty1, ty2)))

let associate_fields fields1 fields2 =
  let rec associate p s s' =
    function
      (l, []) ->
        (List.rev p, (List.rev s) @ l, List.rev s')
    | ([], l') ->
        (List.rev p, List.rev s, (List.rev s') @ l')
    | ((n, t)::r, (n', t')::r') when n = n' ->
        associate ((t, t')::p) s s' (r, r')
    | ((n, t)::r, ((n', t')::_ as l')) when n < n' ->
        associate p ((n, t)::s) s' (r, l')
    | (((n, t)::r as l), (n', t')::r') (* when n > n' *) ->
        associate p s ((n', t')::s') (l, r')
  in let sort = Sort.list (fun (n, _) (n', _) -> n < n') in
  associate [] [] [] (sort fields1, sort fields2)

let rec unify env ty1 ty2 = 
   let expand ty = let res = Ctype.expand_head env ty in
                   if res <> ty then res else raise Not_found
   in
   let rec scan_type ty1 ty2 =
    let ty1 = Ctype.repr ty1 and
        ty2 = Ctype.repr ty2 in
    if ty1 == ty2 then () else
    match (ty1.desc, ty2.desc) with
       (Tvar, _) -> saved_desc := (ty1, ty1.desc) :: !saved_desc;
                    ty2.level <- not_copy_level; (* To check *)
                    ty1.desc <- Tlink ty2
     | (_, Tvar) -> saved_desc := (ty2, ty2.desc) :: !saved_desc;
                    ty1.level <- not_copy_level; (* To check *)
                    ty2.desc <- Tlink ty1
     | (Tarrow(par1,dst1), Tarrow(par2,dst2)) -> scan_type par1 par2;
                                                 scan_type dst1 dst2
     | (Ttuple(types1), Ttuple(types2)) -> List.iter2 scan_type types1 types2
     | (Tconstr(decl1, types1,_), Tconstr(decl2, types2,_)) when
         Path.same decl1 decl2 ->
            List.iter2 scan_type types1 types2 
               (* Careful ! With functors, we may instanciate abstract
                  types - This not really a pb here
                  since inlining is intra-module only, so
                  we cannot inline code from a functor out of it *)
     | (Tnil, _) | (_, Tnil) | (Tfield _, _) | (_, Tfield _) ->
           Message.error "Simplify.unify" (fun () -> print_string 
           "object field cannot be in place of a type")
     | (Tobject (fi1, nm1), Tobject (fi2, nm2)) ->
        unify_fields env fi1 fi2;
        begin match !nm2 with
          Some (_, va::_) when (Ctype.repr va).desc = Tvar -> ()
        | _                                          -> nm2 := !nm1
        end
     | (_, _) -> try
          scan_type (expand ty1) ty2
       with Not_found -> try
          scan_type ty1 (expand ty2)
       with Not_found -> failwith ""
      in try
            scan_type ty1 ty2
         with Failure s ->
            Message.error "unify"
		 (fun () -> print_string s;
                            Printtyp.type_expr ty1;
                            print_string " and ";
                            Printtyp.type_expr ty2;
                            print_string " do not correspond.")

and unify_fields env ty1 ty2 = 
  let (fields1, rest1) = flatten_fields ty1
  and (fields2, rest2) = flatten_fields ty2 in
  let (pairs, miss1, miss2) = associate_fields fields1 fields2 in
  let va = Ctype.newvar () in
  unify env rest1 (build_fields miss2 va);
  unify env (build_fields miss1 va) rest2;
  List.iter (fun (t1, t2) -> unify env t1 t2) pairs


(* Restored type descriptions *)

let cleanup_types () =
  List.iter (fun (ty, desc) -> 
               (match ty.desc with
                  Tlink t -> t.level <- generic_level
                | _ -> ());
               ty.desc <- desc) !saved_desc;
  saved_desc := []

let abbreviations = ref (ref Mnil)
  (* Abbreviation memorized. *)

let rec copy ty =
  let ty = Ctype.repr ty in
  if ty.level == not_copy_level then
    ty
  else begin
    let desc = ty.desc in
    saved_desc := (ty, desc)::!saved_desc;
    let t = Ctype.newvar () in              (* Stub *)
    t.level <- not_copy_level;
    ty.desc <- Tlink t;
    t.desc <-
      begin match desc with
        Tvar ->
          Tvar
      | Tarrow (t1, t2) ->
          Tarrow (copy t1, copy t2)
      | Ttuple tl ->
          Ttuple (List.map copy tl)
      | Tconstr (p, tl, _) ->
          (*
             One must allocate a new reference, so that abbrevia-
             tions belonging to different branches of a type are
             independent.
             Moreover, a reference containing a [Mcons] must be
             shared, so that the memorized expansion of an abbrevi-
             ation can be released by changing the content of just
             one reference.
          *)
          Tconstr (p, List.map copy tl,
                   ref (match ! !abbreviations with
                          Mcons _ -> Mlink !abbreviations
                        | abbrev  -> abbrev))
      | Tobject (t1, {contents = name}) ->
          let name' =
            match name with
              None ->
                None
            | Some (p, tl) ->
                Some (p, List.map copy tl)
          in
          Tobject (copy t1, ref name')
      | Tfield (label, t1, t2) ->
          Tfield (label, copy t1, copy t2)
      | Tnil ->
          Tnil
      | Tlink t -> (* Actually unused *)
          Tlink (copy t)
      end;
    t
  end


(* Renames an expression *)

let rec renameidents table (exp : Typedtree.expression) =
(* table : new names of idents, to avoid having two different idents 
     for the same name
   exp : the exp to rename *)
    let indesc desc = 
       {exp_desc = desc; 
        exp_loc = exp.exp_loc; 
        exp_type = copy exp.exp_type;
        exp_env = exp.exp_env
   } in
         match exp.exp_desc with
             Texp_ident(Path.Pident i, v) ->
                (try
                   indesc (Texp_ident(Path.Pident(Ident.find_same i table), 
                    {val_type = copy v.val_type; val_kind = v.val_kind}))
                with
                   Not_found -> indesc exp.exp_desc)
           | Texp_ident(_,_) -> indesc exp.exp_desc
           | Texp_constant(_) -> indesc exp.exp_desc
           | Texp_let(rec_flag, stack_flag, bindings, body) -> 
                let (table', bindings') = createtablebindings table bindings in
                indesc(Texp_let(rec_flag, stack_flag, renamebindings 
                   (if rec_flag = Recursive then table' else table) bindings',
                   renameidents table' body))
           | Texp_function(clauses) -> 
                indesc (Texp_function(List.map (fun (p, e) ->
                   let (table', p') = renamepat table p in
                      (p', renameidents table' e)) clauses))
           | Texp_apply(rator, rands) -> 
                 indesc(Texp_apply(renameidents table rator,
                 List.map (renameidents table) rands))
           | Texp_match(body, clauses) -> 
                indesc(Texp_match(renameidents table body,
                 List.map (fun (p, e') ->
                 let (table', p') = renamepat table p in
                     (p', renameidents table' e')) clauses))
           | Texp_try(body,clauses) -> indesc(Texp_try(renameidents table body,
                 List.map (fun (p, e') ->
                 let (table', p') = renamepat table p in
                     (p', renameidents table' e')) clauses))
           | Texp_tuple(rands) -> 
                indesc(Texp_tuple(List.map (renameidents table) rands))
           | Texp_construct(c, rands) -> 
                indesc(Texp_construct(c, List.map 
                   (renameidents table) rands))
           | Texp_record(fieldlist) -> let (fl, el) = List.split fieldlist in
                indesc(Texp_record(List.combine fl 
                   (List.map (renameidents table) el)))
           | Texp_field(f, lbl) -> 
                indesc(Texp_field(renameidents table f, lbl))
           | Texp_setfield(e1, lbl, e2) -> 
                indesc(Texp_setfield(renameidents table e1, lbl, 
                                     renameidents table e2))
           | Texp_array(rands) ->  
                indesc(Texp_array(List.map (renameidents table) rands))
           | Texp_ifthenelse(cond, e1,e2) -> 
                indesc(Texp_ifthenelse(
                  renameidents table cond, 
                  renameidents table e1, 
                  match e2 with
                  None -> None
                | Some e3 -> Some (renameidents table e3)))
           | Texp_sequence(e1, e2) -> 
                  indesc(Texp_sequence(renameidents table e1, 
                   renameidents table e2))
           | Texp_while(cond, body) -> 
                  indesc(Texp_while(renameidents table cond, 
                   renameidents table body))
           | Texp_for(i, first, last, dir, body) -> 
                  let x = Ident.create (Ident.name i) in
                  indesc(Texp_for(x, renameidents table first, 
                                     renameidents table last, dir, 
                                     renameidents (Ident.add i x table) body))
           | Texp_when(cond, body) ->
                  indesc(Texp_when(renameidents table cond, 
                                   renameidents table body))
           | Texp_send(exp,m) -> indesc(Texp_send(renameidents table exp,m))
           | Texp_new _ -> indesc exp.exp_desc
           | Texp_instvar _ -> indesc exp.exp_desc
           | Texp_setinstvar (a,b,exp) -> 
                  indesc (Texp_setinstvar(a,b,renameidents table exp))
           | Texp_override (a, rands) -> 
                  indesc (Texp_override(a, List.map
                     (fun (p,e) -> (p, renameidents table e)) rands))
 
and createtablebindings table = function [] -> (table, [])
    | ((p,e)::l) -> let (table', p') = renamepat table p in
             let (table'', bind') = createtablebindings table' l in
                 (table'', (p',e)::bind')

and renamebindings table = List.map 
   (fun (p,e) -> (p, renameidents table e))

and renamepat table pat =
    let indescpat p = {pat_desc = p; pat_loc = pat.pat_loc; pat_type = copy pat.pat_type} in
    match pat.pat_desc with
       Tpat_any -> (table, pat)
     | Tpat_var(i) -> let x = Ident.create (Ident.name i) in
          (Ident.add i x table, indescpat (Tpat_var x))
     | Tpat_alias(p, i) -> let (table', p') = renamepat table p in
          let x = Ident.create (Ident.name i) in
          (Ident.add i x table', indescpat (Tpat_alias(p',x)))
     | Tpat_constant(_) -> (table, pat)
     | Tpat_tuple(patlist) -> let (table', pl') = renamepatlist table patlist
          in (table', indescpat(Tpat_tuple(pl')))
     | Tpat_construct(c, patlist) -> let (table', pl') = renamepatlist table patlist in (table', indescpat(Tpat_construct(c, pl')))
     | Tpat_record(fieldlist) -> let (fl, pl) = List.split fieldlist in
         let (table', pl') = renamepatlist table pl in
         (table', indescpat(Tpat_record(List.combine fl pl')))
     | Tpat_or(_,_) -> (table, pat) (* An or pattern cannot declare variables *)

and renamepatlist table = function
   [] -> (table, [])
 | (p::l) -> let (table', p') = renamepat table p in
             let (table'', l') = renamepatlist table' l in
               (table'', p'::l')

let renamemain patlist body =
    let (table, patlist') = renamepatlist Ident.empty patlist in
    let res = (patlist', renameidents table body) in
       cleanup_types();
       res


let rec parse_function n exp = if n=0 then ([], exp) else
         match exp.exp_desc with
            Texp_ident(Path.Pident i,v) -> 
              if !Control.inlining_sup <= 0 then ([],exp) else
              (try
                 let (taille, rank, expr) = Ident.find_same i (!inline_sup_info) in
		    if rank > n then ([], exp) else
                    begin
                       Message.debug 3 (fun () -> 
                          print_string "Inlining higher order function ";
                          Printtyp.ident i);
                       (* We must instanciate the expression expr so
                          that it has the type in which it is used =>
                            Ctype.unify exp.exp_env exp.exp_type expr.exp_type;
                          but this does not work because the level of the types
                          are changed at the end of typing !
                        *)
                       unify exp.exp_env exp.exp_type expr.exp_type;
                       let (pat', body') = parse_function n expr in
                                renamemain pat' body'
                    end
              with Not_found -> ([],exp))
         |  Texp_function([pattern,body]) ->
                 let (patterns, body') = parse_function (n-1) body in
                    (pattern :: patterns, body')
         |  Texp_function(((pat1, body1)::_) as clauses) ->
                 let pat_type = pat1.pat_type
                 and x = Ident.create "tmp" in
                 let x_as_path = Path.Pident x in
                 let value_description = {val_type=pat_type;
                                          val_kind=Val_reg}
                 in let x_as_exp={exp_loc = Location.none;
                                  exp_type = pat_type;
                                  exp_desc = Texp_ident(x_as_path,
                                               value_description);
                                  exp_env = exp.exp_env} in
                 let x_as_pattern={pat_loc = Location.none;
                                   pat_type = pat_type;
                                   pat_desc = Tpat_var(x)} in
                 let body_as_exp = {exp_loc = Location.none;
                                    exp_type = body1.exp_type;
                                    exp_desc = Texp_match(x_as_exp, clauses);
                                    exp_env = exp.exp_env}
                 in ([x_as_pattern], body_as_exp)
          |  Texp_let(rec_flag, stack_flag, bindings, body) ->
                 let (patterns, body') = parse_function n body in
                     (patterns, {exp_loc = exp.exp_loc;
                                 exp_type = body'.exp_type;
                                 exp_desc = Texp_let(rec_flag, stack_flag, bindings, body');
                                 exp_env = exp.exp_env})
          |  _ -> ([], exp)

(* Transfo 
     let x = M in N   => N[x -> M]
   if M is an ident (avoids losing information when x is a function), 
   or M is a function declaration and x appears only once (or no time) in N
   (avoids creating a closure)
*)

exception Too_many_uses

let count_uses ident e = 
    let rec ct usesmax e = match e.exp_desc with
   Texp_ident(Path.Pident(i),_) when Ident.same i ident ->
         if (usesmax<1) then raise Too_many_uses else usesmax-1
 | Texp_ident(_,_) -> usesmax
 | Texp_constant(_) -> usesmax
 | Texp_let(_,_, bindings, body) -> ct (ctbind usesmax bindings) body
 | Texp_function(clauses) -> ctbind usesmax clauses
 | Texp_apply(rator, rands) -> ct (List.fold_left ct usesmax rands) rator
 | Texp_match(body, clauses) -> ct (ctbind usesmax clauses) body
 | Texp_try(body, clauses) -> ct (ctbind usesmax clauses) body
 | Texp_tuple(rands) -> List.fold_left ct usesmax rands
 | Texp_construct(_, rands) -> List.fold_left ct usesmax rands
 | Texp_record(fielddesc) -> let (_, rands) = List.split fielddesc in
      List.fold_left ct usesmax rands
 | Texp_field(body, _) -> ct usesmax body
 | Texp_setfield(lhs, _, rhs) -> ct (ct usesmax lhs) rhs
 | Texp_array(rands) -> List.fold_left ct usesmax rands
 | Texp_ifthenelse(e1, e2, e3) -> let t = ct (ct usesmax e1) e2 in
      (match e3 with
          None -> t
        | Some e -> ct t e)
 | Texp_sequence(e1, e2) -> ct (ct usesmax e1) e2
 | Texp_while(cond, body) -> ct (ct usesmax cond) body
 | Texp_for(_, first, last, _, body) -> ct (ct (ct usesmax first) last) body
 | Texp_when(cond, body) -> ct (ct usesmax cond) body 
 | Texp_send(exp,_) -> ct usesmax exp
 | Texp_new _ -> usesmax
 | Texp_instvar _ -> usesmax
 | Texp_setinstvar (_,_,exp) -> ct usesmax exp
 | Texp_override (_, rands) -> List.fold_left ct usesmax (List.map snd rands)

and ctbind usesmax = function
   [] -> usesmax
 | ((pat, body)::bindings') -> ctbind (ct usesmax body) bindings'

in try
     1 - (ct 1 e)
   with Too_many_uses -> 2

let substitute ident substexp e = 
    let rec subst exp = 
    let indesc desc = 
       {exp_desc = desc; 
        exp_loc = exp.exp_loc; 
        exp_type = exp.exp_type;
        exp_env = exp.exp_env
   } in
         match exp.exp_desc with
             Texp_ident(Path.Pident i, v) when Ident.same ident i ->
                unify exp.exp_env exp.exp_type substexp.exp_type; 
                let res = renameidents Ident.empty substexp in
		(* In fact, it is not necessary to rename the idents,
                   but we must instanciate the types *)
                cleanup_types();
                res
           | Texp_ident(_,_) -> exp
           | Texp_constant(_) -> exp
           | Texp_let(rec_flag, stack_flag, bindings, body) -> 
                indesc(Texp_let(rec_flag, stack_flag, List.map (fun (p, e') ->
                     (p, subst e')) bindings, subst body))
           | Texp_function(clauses) -> 
                indesc (Texp_function(List.map (fun (p, e) ->
                      (p, subst e)) clauses))
           | Texp_apply(rator, rands) -> 
                 indesc(Texp_apply(subst rator,
                 List.map subst rands))
           | Texp_match(body, clauses) -> 
                indesc(Texp_match(subst body,
                 List.map (fun (p, e') -> (p, subst e')) clauses))
           | Texp_try(body,clauses) -> indesc(Texp_try(subst body,
                 List.map (fun (p, e') -> (p, subst e')) clauses))
           | Texp_tuple(rands) -> 
                indesc(Texp_tuple(List.map subst rands))
           | Texp_construct(c, rands) -> 
                indesc(Texp_construct(c, List.map subst rands))
           | Texp_record(fieldlist) -> let (fl, el) = List.split fieldlist in
                indesc(Texp_record(List.combine fl 
                   (List.map subst el)))
           | Texp_field(f, lbl) -> 
                indesc(Texp_field(subst f, lbl))
           | Texp_setfield(e1, lbl, e2) -> 
                indesc(Texp_setfield(subst e1, lbl, subst e2))
           | Texp_array(rands) ->  
                indesc(Texp_array(List.map subst rands))
           | Texp_ifthenelse(cond, e1,e2) -> 
                indesc(Texp_ifthenelse(
                  subst cond, 
                  subst e1, 
                  match e2 with
                  None -> None
                | Some e3 -> Some (subst e3)))
           | Texp_sequence(e1, e2) -> 
                  indesc(Texp_sequence(subst e1, subst e2))
           | Texp_while(cond, body) -> 
                  indesc(Texp_while(subst cond, subst body))
           | Texp_for(i, first, last, dir, body) -> 
                  indesc(Texp_for(i, subst first, subst last, dir, 
                                     subst body))
           | Texp_when(cond, body) ->
                  indesc(Texp_when(subst cond, subst body))
           | Texp_send(exp,m) -> indesc(Texp_send(subst exp,m))
           | Texp_new _ -> exp
           | Texp_instvar _ -> exp
           | Texp_setinstvar (a,b,exp) -> 
                  indesc (Texp_setinstvar(a,b,subst exp))
           | Texp_override (a, rands) -> 
                  indesc (Texp_override(a, List.map
                     (fun (p,e) -> (p, subst e)) rands))
   in subst e 

(* Simplificateur *)

let rec simplify exp =
   match exp.exp_desc with
  Texp_let(Nonrecursive, _, [({pat_desc=Tpat_var(ident)}, init)],
    {exp_desc=Texp_ident(Path.Pident(ident'), _)}) when ident=ident' ->  
      simplify init

| Texp_let(_, _, [({pat_desc=Tpat_var(ident)}, 
   ({exp_desc=Texp_ident(Path.Pident ident', _)} as init))], body) 
   when (ident<>ident') && (!Control.inlining_sup > 0)  ->
      simplify (substitute ident init body)

| Texp_let(Nonrecursive, _, [({pat_desc=Tpat_var(ident)}, 
   ({exp_desc=Texp_function(_)} as init))], body) 
   when (!Control.inlining_sup > 0) ->
    (match count_uses ident body with
      0 -> simplify body
    | 1 -> simplify (substitute ident init body)
    | _ -> exp)

| Texp_let(Nonrecursive, st, ((a::b::l) as bindings), body) ->
    simplify (List.fold_right (fun x accu -> 
      {exp_desc = Texp_let(Nonrecursive, st, [x], accu);
       exp_loc = exp.exp_loc;
       exp_type = exp.exp_type;
       exp_env = exp.exp_env}) bindings body)

| Texp_apply({exp_desc = Texp_let(rec_flag, stack_flag, bindings, body)} as rator, rands) ->
    simplify {exp_desc = Texp_let(rec_flag, stack_flag, bindings, 
       {exp_desc = Texp_apply(body, rands);
        exp_loc = exp.exp_loc;
        exp_type = exp.exp_type;
        exp_env = exp.exp_env});
       exp_loc = exp.exp_loc;
       exp_type = exp.exp_type;
       exp_env = exp.exp_env}

| Texp_apply(rator, rands) ->
    (* Here, rator contains no let at the top, so we are sure that if
       parse_function succeeds to find patterns, rator begins with fun,
       so the parameters must be evaluated before rator *)
    let nrand = List.length rands in
    let (pat', body') = parse_function nrand rator in
    let npat = List.length pat' in
        if (npat > 0) & (npat < nrand) then
           let (bindings, args) = list_combine_add pat' rands in
                    simplify {exp_desc = Texp_let(Nonrecursive, Anystack,
                          List.rev bindings, {exp_desc = Texp_apply(body',args);
                                              exp_loc = exp.exp_loc;
                                              exp_type = exp.exp_type;
                                              exp_env = exp.exp_env});
                     exp_loc = exp.exp_loc;
                     exp_type = exp.exp_type;
                     exp_env = exp.exp_env}  
        else if npat = nrand then
                    simplify {exp_desc = Texp_let(Nonrecursive, Anystack,
                          List.rev (List.combine pat' rands), body');
                     exp_loc = exp.exp_loc;
                     exp_type = exp.exp_type;
                     exp_env = exp.exp_env}
        else exp

| _ -> exp

(* Select functions which will perhaps be inlined *)

let rec functional ty = match ty.desc with
   Tarrow _ -> true
 | Tlink ty -> functional ty
 | _ -> false

let rec higher_order_rank e =
   match e.exp_desc with
      Texp_function([pat,body]) -> if functional pat.pat_type then 1 
                                   else (higher_order_rank body)+1
    | _ -> raise Not_higher_order

let rec is_compatible1 pat exp = match (pat.pat_desc,exp.exp_desc) with
    (Tpat_any, _) -> ()
  | (Tpat_var x, Texp_ident(Path.Pident(y),_)) when x = y -> ()
  | (Tpat_alias (p, x), Texp_ident(Path.Pident(y),_)) -> 
       if x <> y then is_compatible1 p exp
  | _ -> raise Not_higher_order (* This could be improved : we should
check that the expression exp matches the pattern pat *)

let rec is_compatible pats exps = match (pats, exps) with
    ([], _) -> exps
  | (pat :: lpat, exp :: lexps) -> is_compatible1 pat exp; is_compatible lpat lexps

(* Transforms alls applications Texp_apply(x, pats) into xnew and
checks that there is no other call to x *)
let check_calls x pats xnew e = 
  let rec ck e =
   match e.exp_desc with
   Texp_ident(Path.Pident y,_) when x = y -> raise Not_higher_order
 | Texp_ident(_,_) -> e
 | Texp_constant(_) -> e
 | Texp_let(r,s, bindings, body) -> indesc e (Texp_let(r,s,ckbind bindings, ck body))
 | Texp_function(clauses) -> indesc e (Texp_function(ckbind clauses))
 | Texp_apply(rator, rands) -> begin
    match rator.exp_desc with
        Texp_ident(Path.Pident y,_) when x = y -> 
          begin
            match is_compatible pats rands with
              [] -> xnew
            | lrands -> indesc e (Texp_apply (xnew, List.map ck lrands))
          end
      | _ -> indesc e (Texp_apply(ck rator, List.map ck rands))
    end
 | Texp_match(body, clauses) -> indesc e (Texp_match(ck body, ckbind clauses))
 | Texp_try(body, clauses) -> indesc e (Texp_try (ck body, ckbind clauses))
 | Texp_tuple(rands) -> indesc e (Texp_tuple(List.map ck rands))
 | Texp_construct(c, rands) -> indesc e (Texp_construct(c,List.map ck rands))
 | Texp_record(fielddesc) -> indesc e (Texp_record(List.map (fun (pat, e) -> (pat,ck e)) fielddesc))
 | Texp_field(body, f) -> indesc e (Texp_field(ck body,f))
 | Texp_setfield(lhs, f, rhs) -> indesc e (Texp_setfield(ck lhs, f, ck rhs))
 | Texp_array(rands) -> indesc e (Texp_array(List.map ck rands))
 | Texp_ifthenelse(e1, e2, e3) -> indesc e (Texp_ifthenelse(ck e1, ck e2, 
      (match e3 with
          None -> None
        | Some e -> Some(ck e))))
 | Texp_sequence(e1, e2) -> indesc e (Texp_sequence (ck e1, ck e2))
 | Texp_while(cond, body) -> indesc e (Texp_while (ck cond, ck body))
 | Texp_for(i, first, last, dir, body) -> indesc e (Texp_for(i, ck first, ck last, dir, ck body))
 | Texp_when(cond, body) -> indesc e (Texp_when (ck cond, ck body))
 | Texp_send(exp,m) -> indesc e (Texp_send(ck exp,m))
 | Texp_new _ -> e
 | Texp_instvar _ -> e
 | Texp_setinstvar (a,b,exp) -> indesc e (Texp_setinstvar (a,b,ck exp))
 | Texp_override (a, rands) -> indesc e (Texp_override (a,List.map (fun (pat, e) -> (pat,ck e)) rands))
and ckbind l = List.map (fun (pat, e) -> (pat,ck e)) l
in ck e

let rec is_higher_order_rec x pats rank e =
   if rank = 0 then 
      let xnew = Ident.create "tmp" in
      let xnew_ident = {exp_loc = Location.none;
                        exp_type = e.exp_type;
                        exp_desc = Texp_ident(Path.Pident xnew, 
                                     {val_type = e.exp_type;
                                      val_kind = Val_reg});
                        exp_env = e.exp_env} in
      let xnew_pat = {pat_loc = Location.none;
                      pat_type = e.exp_type;
                      pat_desc = Tpat_var(xnew)} in
          indesc e (Texp_let(Recursive, Anystack, 
          [xnew_pat, check_calls x (List.rev pats) xnew_ident e], xnew_ident))
   else match e.exp_desc with
      Texp_function([pat,body]) -> indesc e (Texp_function([pat,
      is_higher_order_rec x (pat::pats) (rank-1) body]))
    | _ -> raise Not_higher_order

let put_inline_info info rec_flag (pat, e) =
   match (pat.pat_desc, e.exp_desc) with
      (Tpat_var(x), Texp_function([_, _])) -> 
            (try
              let rank = higher_order_rank e in
              let e = match rec_flag with
                        Nonrecursive -> e
                      | Recursive -> is_higher_order_rec x [] rank e
              in
              let taille1 = !Control.inlining_sup - (taille (!Control.inlining_sup) e) in
              Message.debug 3 (fun () -> 
                   print_string "Higher order function ";
                   Printtyp.ident x;
                   print_string " can be inlined (size : ";
                   print_int taille1;
                   print_string ")");
              Ident.add x  (taille1, rank, e) info
            with
              Too_big -> Message.debug 4 (fun () -> 
                            print_string "Higher order function ";
                            Printtyp.ident x;
                            print_string " cannot be inlined because too big (size > ";
                            print_int (!Control.inlining_sup);
                            print_string ")");
                         info
            | Not_higher_order -> info)
    | _ -> info

